
// Operations on irregular block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `IrregRangeRep': irregular block-distributed template range.  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

IrregRangeRep :: IrregRangeRep(RangeRep* _kernel, int* _blocks) :
	RangeRep(0, _kernel, _kernel->lev() + 1, _kernel->dim(),
                 DIST_IRREG) {
  kernel->incr() ;

  nBlocks = kernel->size() ;
  map     = new int [nBlocks] ;
  blocks  = new int [nBlocks] ;
  blockSize = 0 ;

  for(int i = 0 ; i < nBlocks ; i++) {
    const int size = _blocks [i] ;
    
    blocks [i] = size ;
    map    [i] = extent ;

    if(size > blockSize) blockSize = size ;
    extent += size ;
  }

  if(extent) scale = ((float) nBlocks) / extent ;
}

IrregRangeRep :: IrregRangeRep(const int _extent, RangeRep* _kernel,
                               int* _map) :
	RangeRep(_extent, _kernel, _kernel->lev() + 1, _kernel->dim(),
                 DIST_IRREG) {

  // Support for Global Arrays

  kernel->incr() ;

  nBlocks = kernel->size() ;
  map     = new int [nBlocks] ;
  blocks  = new int [nBlocks] ;
  blockSize = 0 ;

  if(extent) scale = ((float) nBlocks) / extent ;

  const int lb = _map [0] ;

  int curr = lb ;
  for(int i = 0 ; i < nBlocks ; i++) {
    const int ip   = i + 1 ;
    const int next = ip < nBlocks ? _map [ip] : lb + extent ;

    if(next < curr)
      AdlibError("IrregRangeRep :: IrregRangeRep: "
                 "_map vector not monotonically increasing\n") ;

    const int size = next - curr ;
    
    blocks [i] = size ;
    map    [i] = curr - lb ;

    if(size > blockSize) blockSize = size ;

    curr = next ;
  }
}

IrregRangeRep :: ~IrregRangeRep() {
  delete [] blocks ;
  delete [] map ;

  kernel->decr() ;
}

RangeRep* IrregRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

RangeRep* IrregRangeRep :: subrng(const int _extent,
                                  const int _base, const int _stride) {
  return new IrregSubRangeRep(_extent, _base, _stride, this, kernel, this) ;
}

void IrregRangeRep :: location(Location* loc, const int glb) const {

  loc->tem = glb ;
  loc->crd = findBlock(glb) ;
  loc->sub = glb - map [loc->crd] ;
  loc->dim = dimension ;
}

void IrregRangeRep :: block(const int ker_glb, Block* blk) const {

  // `ker_glb' is a global subscript in the kernel range.

  blk->tem_stp = 1 ;
  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  blk->sub_bas = 0 ;
  blk->tem_bas = map [ker_glb] ;
  blk->glb_bas = blk->tem_bas ;

  blk->count   = blocks [ker_glb] ;
}

int IrregRangeRep :: volume() const {
  return blockSize ;
}

void IrregRangeRep :: show() {
  AdlibPrintf("IrregRangeRep :: show: size = %d\n", extent) ; 
}

// Obsolescent stuff ...

int IrregRangeRep :: glb(const int sub, const int ker_glb) const {
  return map [ker_glb] + sub ;
}

MEMORY_POOL_DEFN(IrregRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `IrregSubRangeRep': irregular block-distributed subrange.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

IrregSubRangeRep :: IrregSubRangeRep(const int _extent, const int _base, const int _stride,
		                     RangeRep* _parent, RangeRep* _kernel, IrregDist* dist) :
	SubRangeRep(_extent, _base, _stride, _parent,
		    _kernel, _kernel->lev() + 1, _kernel->dim(), DIST_IRREG),
	IrregDist(*dist) {

  if(stride > 0) {
    ker_lo = findBlock(base) ;
    ker_hi = findBlock(base + stride * (extent - 1)) ;
  }
  else {
    ker_hi = findBlock(base) ;
    ker_lo = findBlock(base + stride * (extent - 1)) ;
  }
}

RangeRep* IrregSubRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

RangeRep* IrregSubRangeRep :: subrng(const int _extent,
                                    const int _base, const int _stride) {
  return new IrregSubRangeRep(_extent,
			      base + stride * _base, stride * _stride,
			      parent, kernel, this) ;
}

void IrregSubRangeRep :: location(Location* loc, const int glb) const {
  const int tem = base + stride * glb ;

  loc->tem = tem ;
  loc->crd = findBlock(tem) ;
  loc->sub = tem - map [loc->crd] ;
  loc->dim = dimension ;
}

void IrregSubRangeRep :: block(const int ker_glb, Block* blk) const {

  blk->tem_stp = stride ;
  blk->sub_stp = stride ;
  blk->glb_stp = 1 ;

  if(ker_glb >= ker_lo && ker_glb <= ker_hi) {
    const int blockStart        = map [ker_glb] ;
    const int blockStartOffset  = blockStart - base ;
    const int blockFinishOffset = blockStartOffset + blocks [ker_glb] - 1 ;

    int i_lo, i_hi ;

    if(stride > 0) {
      i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    }
    else {
      i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    }

    blk->tem_bas = base + stride * i_lo ;
    blk->sub_bas = blk->tem_bas - blockStart ;
    blk->glb_bas = i_lo ;

    blk->count   = i_hi - i_lo + 1 ;
  }
  else {
    blk->tem_bas = 0 ;
    blk->sub_bas = 0 ;
    blk->glb_bas = 0 ;

    blk->count   = 0 ;
  }
}

RangeRep* IrregSubRangeRep :: subker() const {

  int ker_bas, ker_str ;
  if(stride > 0) {
    ker_bas = ker_lo ;
    ker_str = 1 ;
  }
  else {
    ker_bas = ker_hi ;
    ker_str = -1 ;
  }

  return kernel->subrng(ker_hi - ker_lo + 1, ker_bas, ker_str) ;
}

int IrregSubRangeRep :: volume() const {
  return blockSize ;
}

void IrregSubRangeRep :: show() {
  AdlibPrintf("IrregSubRangeRep :: show: size = %d base = %d stride = %d\n",
              extent, base, stride) ;
}

// Obsolescent stuff ...

int IrregSubRangeRep :: glb(const int sub, const int ker_glb) const {
  return (map [ker_glb] + sub - base) / stride ;
}

MEMORY_POOL_DEFN(IrregSubRangeRep)


// -------------------------------------------------------------
// Things to do
//
//   `subker' returns non-empty subrange even for extent = 0
//        (not a problem?)

